// Cassandra DB Configuration and init
'use strict';

const cassandra = require('cassandra-driver');
const config = require('./config')();
const chalk = require('chalk');
const glob = require('glob');
const winston = require('winston');
const cluster = require('cluster');
const fs = require('fs');
const path = require('path');

// Print Once
function _printOnce(){
  return cluster.isMaster || (cluster.worker && cluster.worker.id==1);
}

// Connect to DB
function connect(options){
  var Servers, Keyspace;
  var cassandraOptions = {};
  if(config.get().bigdata){
    if(config.get().bigdata.servers){
      Servers = config.get().bigdata.servers;
    }

    if(config.get().bigdata.keyspace){
      Keyspace = config.get().bigdata.keyspace;
    }
  }

  if(options && options.servers){
    Servers = options.servers;
    if(options.keyspace){
      Keyspace = options.keyspace;
    }
  }

  if(Servers){
    cassandraOptions.contactPoints = Servers;
  }
  if(Keyspace){
    Keyspace = Keyspace.replace(/-/g,'_');
    cassandraOptions.keyspace = Keyspace;
  }

  var client=new cassandra.Client(cassandraOptions);
  client.connect(function(err){
    if(err){
      winston.error('Error with BigData DataBase', Servers, 'and keyspace' , Keyspace, ':', err);
      process.exit();
    }else{
      if(_printOnce()){
        winston.info('-'.repeat(70));
        winston.info('Using BigData DataBase:', chalk.blue(Servers, 'and keyspace' , Keyspace));
      }
    }
  });
  return client;
}

// Load Queries
function loadQueries(options, client){
  var queries = {};
  var loadedQueries = [];

  var files = glob.sync(global.mio.appPath+'/modules/*/queries/*.cql');
  for (var i = 0; i < files.length; i++) {
    var queryName = path.basename(files[i], '.cql');
    try {
      var _request = fs.readFileSync(files[i], {encoding: 'utf8'});
      queries[queryName+'Src'] = _request;
      if(client){
        queries[queryName] = function(options, cb){
          return client.execute(_request, options, cb);
        };
      }
      if(loadedQueries.indexOf(queryName) == -1) loadedQueries.push(queryName);
    } catch (err) {
      winston.error('Can\'t load query:', queryName);
    }
  }

  return queries;
}

// Magic Init
function init(options){
  var client = connect(options);
  if(client){
    var queries = loadQueries(options, client);
    global.mio.cassandra = {client: client, queries: queries};
    return {client: client, queries: queries};
  }

  return;
}

module.exports = {
  init: init,
  loadQueries: loadQueries,
  connect: connect
};
